from functools import wraps

from django.forms.utils import pretty_name
from django.utils.html import format_html

from jnt_django_toolbox.admin.helpers import urls
from jnt_django_toolbox.helpers.objects import getattr_nested


def admin_link(attr, short_description=None, empty_description="-"):
    """Render a link to a related model in the admin detail page.

    ``attr (str)``
        Name of the related field.
    ``short_description (str)``
        Name if the field.
        Default value: None.
    ``empty_description (str)``
        Value to display if the related field is None.
        Default value: -.

    The wrapped method receives the related object and should
    return the link text.

    Usage::

        from django.contrib import admin
        from jnt_admin_tools.decorators import admin_link
        from test_app.models import Foo

        @admin.register(Foo)
        class FooAdmin(admin.ModelAdmin):
            fields = ('name', 'bar', 'bar_link')
            readonly_fields = ('bar_link',)

            @admin_link('bar')
            def bar_link(self, obj):
                return obj.name

    .. image:: images/decorators/decorator_admin_link.png
    """

    def wrap(func):  # noqa: WPS430
        @wraps(func)
        def field_func(self, isntance):  # noqa: WPS430
            related_obj = getattr_nested(isntance, attr)
            if related_obj is None:
                return empty_description
            url = urls.admin_change_url(related_obj)
            return format_html(
                '<a href="{0}">{1}</a>',
                url,
                func(self, related_obj),
            )

        field_func.short_description = (
            short_description if short_description else pretty_name(attr)
        )
        field_func.allow_tags = True
        return field_func

    return wrap



from django.db import models
from django.db.models import Subquery


class SQCount(Subquery):
    """Count items of subquery."""

    template = "(SELECT count(*) FROM (%(subquery)s) _count)"  # noqa: WPS323
    output_field = models.IntegerField()



import six
from django.contrib.admin import FieldListFilter
from django.contrib.admin.options import IncorrectLookupParameters
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _

from jnt_django_toolbox.models.fields.bit.types import Bit


class BitFieldListFilter(FieldListFilter):
    """List filter for bitfields."""

    def __init__(self, field, request, params, model, model_admin, field_path):
        """Initializing."""
        self.lookup_kwarg = field_path
        self.lookup_val = int(request.GET.get(self.lookup_kwarg, 0))
        self.flags = field.flags
        self.labels = field.labels
        super().__init__(
            field,
            request,
            params,
            model,
            model_admin,
            field_path,
        )

    def queryset(self, request, queryset):
        """Provides queryset."""
        filter = {
            param: models.F(param).bitor(value)
            for param, value in six.iteritems(self.used_parameters)
        }
        try:
            return queryset.filter(**filter)
        except ValidationError as e:
            raise IncorrectLookupParameters(e)

    def expected_parameters(self):
        """Get expected parameters."""
        return [self.lookup_kwarg]

    def choices(self, cl):
        """Get choices."""
        yield {
            "selected": self.lookup_val == 0,
            "query_string": cl.get_query_string({}, [self.lookup_kwarg]),
            "display": _("All"),
        }
        for number, _flag in enumerate(self.flags):
            bit_mask = Bit(number).mask
            yield {
                "selected": self.lookup_val == bit_mask,
                "query_string": cl.get_query_string(
                    {self.lookup_kwarg: bit_mask},
                ),
                "display": self.labels[number],
            }



import inspect
import itertools
from functools import wraps

from django.conf import settings
from opentracing import global_tracer


def trace_span(show_args: bool = False):
    """Wrap function for jaeger tracing."""
    frm = inspect.stack()[1]
    mod = inspect.getmodule(frm[0]).__name__

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):  # noqa: WPS430
            from jnt_django_toolbox.profiling.profilers import JaegerProfiler

            is_tracing = any(
                isinstance(profiler, JaegerProfiler)
                for profiler in settings.REQUEST_PROFILERS
            )
            if not is_tracing:
                return func(*args, **kwargs)

            caller = None
            if args:
                caller = args[0].__class__

            operation_name = "{0}.{1}{2}".format(
                mod,
                "{0}.".format(caller.__name__) if caller else "",
                func.__name__,
            )
            if show_args:
                operation_name = "{0} [{1}]".format(
                    operation_name,
                    ", ".join(
                        itertools.chain(
                            [str(arg) for arg in args],
                            [
                                "{0}={1}".format(kwarg_key, kwarg_value)
                                for kwarg_key, kwarg_value in kwargs.items()
                            ],
                        ),
                    ),
                )

            with global_tracer().start_active_span(
                operation_name,
                child_of=global_tracer().active_span,
            ):
                return func(*args, **kwargs)

        return wrapper

    return decorator



from statistics import mean

from django.db import connection
from django.test.utils import CaptureQueriesContext

from jnt_django_toolbox.profiling.profilers.base import BaseProfiler


class DatabaseQueriesProfiler(BaseProfiler):
    """Database queries count profiler."""

    def __init__(self, header_prefix: str = "app_db"):
        """Initializing."""
        self._header_prefix = header_prefix

    def before_request(self, request, stack):
        """Start capturing requests."""
        self._context = CaptureQueriesContext(connection)
        stack.enter_context(self._context)

    def after_request(self, request, response):
        """Add profiling info to response."""
        query_timings = [float(query["time"]) for query in self._context]
        if query_timings:
            response[self._header("total_time")] = round(sum(query_timings), 3)
            response[self._header("max_time")] = round(max(query_timings), 3)
            response[self._header("avg_time")] = round(mean(query_timings), 3)

        response[self._header("count")] = len(self._context)

    def _header(self, name: str) -> str:
        return "{0}_{1}".format(self._header_prefix, name)



def wrap_cursor(connection):
    """Wrap cursor for logging sql queries."""
    if not hasattr(connection, "_logged_cursor"):  # noqa: WPS421
        connection._logged_cursor = connection.cursor
        connection._logged_chunked_cursor = connection.chunked_cursor
        connection.queries_count = 0

        def cursor(*args, **kwargs):  # noqa: WPS430
            # Per the DB API cursor() does not accept any arguments. There"s
            # some code in the wild which does not follow that convention,
            # so we pass on the arguments even though it"s not clean.
            # See:
            # https://github.com/jazzband/django-debug-toolbar/pull/615
            # https://github.com/jazzband/django-debug-toolbar/pull/896
            return NormalCursorWrapper(
                connection._logged_cursor(*args, **kwargs),
                connection,
            )

        def chunked_cursor(*args, **kwargs):  # noqa: WPS430
            return NormalCursorWrapper(
                connection._logged_chunked_cursor(*args, **kwargs),
                connection,
            )

        connection.cursor = cursor
        connection.chunked_cursor = chunked_cursor
        return cursor


class NormalCursorWrapper:
    """Wraps a cursor and logs queries."""

    def __init__(self, cursor, db):
        """Initialize."""
        self.cursor = cursor
        # Instance of a BaseDatabaseWrapper subclass
        self.db = db

    def callproc(self, procname, params=None):
        """Wrap callproc."""
        return self._record(self.cursor.callproc, procname, params)

    def execute(self, sql, params=None):
        """Wrap execute."""
        return self._record(self.cursor.execute, sql, params)

    def executemany(self, sql, param_list):
        """Wrap executemany."""
        return self._record(self.cursor.executemany, sql, param_list)

    def __getattr__(self, attr):
        return getattr(self.cursor, attr)

    def __iter__(self):
        return iter(self.cursor)

    def __enter__(self):
        return self

    def __exit__(self, type_, value, traceback):
        self.close()

    def _record(self, method, sql, params):
        self.db.queries_count += 1
        return method(sql, params)
